#![allow(dead_code)]
use crate::grpc::brain as grpc;
use dashmap::DashMap;
use std::sync::Arc;
use std::sync::RwLock;
use tokio::sync::mpsc::Sender;
use tokio::sync::oneshot::Sender as OneshotSender;

#[derive(Eq, Hash, PartialEq, Clone)]
pub struct Node {
    pub public_key: String,
    pub owner_key: String,
    pub country: String,
    pub city: String,
    pub ip: String,
    pub avail_mem_mb: u32,
    pub avail_vcpus: u32,
    pub avail_storage_gbs: u32,
    pub avail_ipv4: u32,
    pub avail_ipv6: u32,
    pub avail_ports: u32,
    pub max_ports_per_vm: u32,
}

impl From<grpc::RegisterNodeRequest> for Node {
    fn from(node: grpc::RegisterNodeRequest) -> Self {
        Node {
            public_key: node.node_pubkey,
            owner_key: node.owner_pubkey,
            country: node.country,
            city: node.city,
            ip: node.ip,
            avail_mem_mb: node.avail_memory_mb,
            avail_vcpus: node.avail_vcpus,
            avail_storage_gbs: node.avail_storage_gb,
            avail_ipv4: node.avail_ipv4,
            avail_ipv6: node.avail_ipv6,
            avail_ports: node.avail_ports,
            max_ports_per_vm: node.max_ports_per_vm,
        }
    }
}

impl Into<grpc::NodeListResp> for Node {
    fn into(self) -> grpc::NodeListResp {
        grpc::NodeListResp {
            node_pubkey: self.public_key,
            country: self.country,
            city: self.city,
            ip: self.ip,
            server_rating: 0,
            provider_rating: 0,
        }
    }
}

#[derive(Clone)]
pub struct Contract {
    pub uuid: String,
    pub hostname: String,
    pub admin_pubkey: String,
    pub node_pubkey: String,
    pub exposed_ports: Vec<u32>,
    pub public_ipv4: String,
    pub public_ipv6: String,
    pub disk_size_gb: u32,
    pub vcpus: u32,
    pub memory_mb: u32,
    pub kernel_sha: String,
    pub dtrfs_sha: String,
    pub created_at: String,
}

impl Into<grpc::VmContract> for Contract {
    fn into(self) -> grpc::VmContract {
        grpc::VmContract {
            uuid: self.uuid,
            hostname: self.hostname,
            admin_pubkey: self.admin_pubkey,
            node_pubkey: self.node_pubkey,
            exposed_ports: self.exposed_ports,
            public_ipv4: self.public_ipv4,
            public_ipv6: self.public_ipv6,
            disk_size_gb: self.disk_size_gb,
            vcpus: self.vcpus,
            memory_mb: self.memory_mb,
            kernel_sha: self.kernel_sha,
            dtrfs_sha: self.dtrfs_sha,
            created_at: self.created_at,
        }
    }
}

#[derive(Default)]
pub struct BrainData {
    nodes: RwLock<Vec<Node>>,
    contracts: RwLock<Vec<Contract>>,
    tmp_vmrequests: DashMap<String, grpc::NewVmRequest>,
    cli_vmcontract_tx: DashMap<String, OneshotSender<grpc::NewVmConfirmation>>,
    daemon_deletevm_tx: DashMap<String, Sender<grpc::DeletedVmUpdate>>,
    daemon_newvm_tx: DashMap<String, Sender<grpc::NewVmRequest>>,
}

#[derive(Debug)]
enum TxType {
    CliContract,
    DaemonDeleteVm,
    DaemonNewVm,
}

impl BrainData {
    pub fn new() -> Self {
        Self {
            nodes: RwLock::new(Vec::new()),
            contracts: RwLock::new(Vec::new()),
            tmp_vmrequests: DashMap::new(),
            cli_vmcontract_tx: DashMap::new(),
            daemon_deletevm_tx: DashMap::new(),
            daemon_newvm_tx: DashMap::new(),
        }
    }

    pub fn insert_node(&self, node: grpc::RegisterNodeRequest) {
        let mut nodes = self.nodes.write().unwrap();
        for n in nodes.iter_mut() {
            if n.public_key == node.node_pubkey {
                *n = node.into();
                return;
            }
        }
        nodes.push(node.into());
    }

    pub fn add_daemon_deletevm_tx(
        self: Arc<Self>,
        node_pubkey: &str,
        tx: Sender<grpc::DeletedVmUpdate>,
    ) {
        self.daemon_deletevm_tx.insert(node_pubkey.to_string(), tx);
    }

    pub fn del_daemon_deletevm_tx(&self, node_pubkey: &str) {
        self.daemon_deletevm_tx.remove(&node_pubkey.to_string());
    }

    pub async fn delete_vm(&self, delete_vm: grpc::DeletedVmUpdate) {
        if let Some(contract) = self.find_contract_by_uuid(&delete_vm.uuid) {
            if let Some(daemon_tx) = self.daemon_deletevm_tx.get(&contract.node_pubkey) {
                let _ = daemon_tx.send(delete_vm.clone()).await;
            }
            let mut contracts = self.contracts.write().unwrap();
            contracts.retain(|c| c.uuid != delete_vm.uuid);
        }
    }

    pub fn add_daemon_newvm_tx(self: Arc<Self>, node_pubkey: &str, tx: Sender<grpc::NewVmRequest>) {
        self.daemon_newvm_tx.insert(node_pubkey.to_string(), tx);
    }

    pub fn del_daemon_newvm_tx(&self, node_pubkey: &str) {
        self.daemon_newvm_tx.remove(&node_pubkey.to_string());
    }

    pub async fn submit_vmconfirmation(&self, confirmation: grpc::NewVmConfirmation) {
        let newvmreq = match self.tmp_vmrequests.remove(&confirmation.uuid) {
            Some((_, r)) => r,
            None => return,
        };
        if let Some((_, client_tx)) = self.cli_vmcontract_tx.remove(&newvmreq.admin_pubkey) {
            let _ = client_tx.send(confirmation.clone());
        }
        if confirmation.error == "" {
            return;
        }
        let contract = Contract {
            uuid: confirmation.uuid,
            exposed_ports: confirmation.exposed_ports,
            public_ipv4: confirmation.public_ipv4,
            public_ipv6: confirmation.public_ipv6,
            created_at: format!("{:?}", std::time::SystemTime::now()),
            hostname: newvmreq.hostname,
            admin_pubkey: newvmreq.admin_pubkey,
            node_pubkey: newvmreq.node_pubkey,
            disk_size_gb: newvmreq.disk_size_gb,
            vcpus: newvmreq.vcpus,
            memory_mb: newvmreq.memory_mb,
            kernel_sha: newvmreq.kernel_sha,
            dtrfs_sha: newvmreq.dtrfs_sha,
        };
        self.contracts.write().unwrap().push(contract);
    }

    pub async fn submit_newvmrequest(
        self: Arc<Self>,
        mut req: grpc::NewVmRequest,
        tx: OneshotSender<grpc::NewVmConfirmation>,
    ) -> bool {
        req.uuid = uuid::Uuid::new_v4().to_string();
        self.tmp_vmrequests.insert(req.uuid.clone(), req.clone());
        self.cli_vmcontract_tx
            .insert(req.admin_pubkey.to_string(), tx);
        if let Some(server_tx) = self.clone().daemon_newvm_tx.get(&req.node_pubkey) {
            if server_tx.send(req.clone()).await.is_err() {
                return false;
            }
        }
        true
    }

    pub fn del_cli_vmcontract_tx(&self, admin_pubkey: &str) {
        self.cli_vmcontract_tx.remove(admin_pubkey);
    }

    pub fn insert_contract(&self, contract: Contract) {
        let mut contracts = self.contracts.write().unwrap();
        contracts.push(contract);
    }

    pub fn find_nodes_by_pubkey(&self, public_key: &str) -> Option<Node> {
        let nodes = self.nodes.read().unwrap();
        nodes.iter().cloned().find(|n| n.public_key == public_key)
    }

    pub fn find_ns_by_owner_key(&self, owner_key: &str) -> Option<Node> {
        let nodes = self.nodes.read().unwrap();
        nodes.iter().cloned().find(|n| n.owner_key == owner_key)
    }

    pub fn find_nodes_by_filters(&self, filters: &crate::grpc::brain::NodeFilters) -> Vec<Node> {
        let nodes = self.nodes.read().unwrap();
        nodes
            .iter()
            .cloned()
            .filter(|n| {
                n.avail_ports >= filters.free_ports
                    && (!filters.offers_ipv4 || n.avail_ipv4 > 0)
                    && (!filters.offers_ipv6 || n.avail_ipv6 > 0)
                    && n.avail_vcpus >= filters.vcpus
                    && n.avail_mem_mb >= filters.memory_mb
                    && n.avail_storage_gbs >= filters.storage_gb
            })
            .collect()
    }

    pub fn find_contract_by_uuid(&self, uuid: &str) -> Option<Contract> {
        let contracts = self.contracts.read().unwrap();
        contracts.iter().cloned().find(|c| c.uuid == uuid)
    }

    pub fn find_contracts_by_admin_pubkey(&self, admin_pubkey: &str) -> Vec<Contract> {
        let contracts = self.contracts.read().unwrap();
        contracts
            .iter()
            .cloned()
            .filter(|c| c.admin_pubkey == admin_pubkey)
            .collect()
    }

    pub fn find_contracts_by_node_pubkey(&self, node_pubkey: &str) -> Vec<Contract> {
        let contracts = self.contracts.read().unwrap();
        contracts
            .iter()
            .cloned()
            .filter(|c| c.node_pubkey == node_pubkey)
            .collect()
    }
}
